#include "twoStateModel.h"
#include "matrixUtils.h"
#include "someUtil.h"
#include <cmath>
//#include <cstdlib>
#include <string>
using namespace std;

twoStateModel::twoStateModel(MDOUBLE pi1, MDOUBLE globalRate, MDOUBLE freq1, rootFreqType rootType)
: _rQ(globalRate), _rootType(rootType)
{
	resizeMatrix(_Q, alphabetSize(), alphabetSize());
	_freq.resize(alphabetSize(), 0.0);
	if (_rootType != STATIONARY)
		setFreq(freq1);
	if (_rQ <=0)
		errorMsg::reportError("error in twoStateModel. _rQ must be positive");
	setPi1(pi1); //this also updateQ
}

twoStateModel& twoStateModel::operator=(const twoStateModel &other)
{
	_Q = other._Q;
	_freq = other._freq;
	_rQ = other._rQ;
	_pi1 = other._pi1;
	_rootType = other._rootType;	
	return *this;
}

void twoStateModel::setPi1(const MDOUBLE val)
{ 
	if (val == _pi1)
		return;
	if (!pijt_is_prob_value(val))
		errorMsg::reportError("Error in twoStateModel::setPi1, pi1 <0 or >1");
    _pi1 = val; 
	updateQ();
	return;
}

void twoStateModel::setFreq(MDOUBLE f1)
{
	if (_rootType == STATIONARY)
		errorMsg::reportError("twoStateModel::setFreq. Cannot set frequencies when determined by the stationary distribution");
	if (!pijt_is_prob_value(f1))
		errorMsg::reportError("Error in twoStateModel::setFreq, f1 <0 or >1");
	_freq[0] = 1 - f1;
	_freq[1] = f1;
}



//this function doesn't updateQ as the global rate is accounted for in the Pijt calculations
void twoStateModel::setScaleRate(const MDOUBLE val)
{ 
	if (val == _rQ)
		return;
	if (val <=0)
		errorMsg::reportError("error in twoStateModel. _rQ must be positive");
    _rQ = val; 
	updateQ();
	return;
}


/********************************************************************************************
*********************************************************************************************/
void twoStateModel::updateQ()
{
	_Q[0][1] = _pi1;
	_Q[0][0] = -_Q[0][1];
	_Q[1][0] = 1 - _pi1; 
	_Q[1][1] = -_Q[1][0]; 
	for (int i=0; i<_Q.size();i++) {
		MDOUBLE sum = _Q[i][0]+_Q[i][1];
		if ((std::abs(sum)>err_allow_for_pijt_function()))
			errorMsg::reportError("Error in twoStateModel::updateQ, sum of row is not 0");
	}
	if (_rootType == STATIONARY)
	{
		_freq[0] = _Q[1][0];
		_freq[1] = _Q[0][1];
	}
	_Q = multiplyMatrixByScalar(_Q, _rQ);

	//normalize matrix???
}
/********************************************************************************************
*********************************************************************************************/
const MDOUBLE twoStateModel::freq(const int i) const {
	if (i >= _freq.size()) 
		errorMsg::reportError("Error in twoStateModel::freq, i > size of frequency vector");
	return _freq[i];
}


/********************************************************************************************
Pij_t - Based on Analytic solution
*********************************************************************************************/
const MDOUBLE twoStateModel::Pij_t(const int i,const int j, const MDOUBLE t1)  const {
	//MDOUBLE t = t1 * _rQ;
	MDOUBLE t = t1;;

	MDOUBLE gain = _Q[0][1];
	MDOUBLE loss = _Q[1][0];
	MDOUBLE eigenvalue =  -(gain + loss);
	
	int caseNum = i + j*2;
	MDOUBLE res = -1;
	switch (caseNum) 
	{
		case 0 : 
			res = loss/(-eigenvalue) + exp(eigenvalue*t)*(1 - loss/(-eigenvalue)); 
			//in case the matrix is normalized can be calculated according to Huelsenbeck, nielsen and bollback
			//res = pi0 + pi1*exp(-m*t) where m = 1/(2pi1pi0)
			break;
		case 1 : 
			res =  loss/(-eigenvalue) - exp(eigenvalue*t)*(1 - gain/(-eigenvalue)); 
			break;
		case 2 : 
			res =  gain/(-eigenvalue) - exp(eigenvalue*t)*(1 - loss/(-eigenvalue));
            break;
		case 3 : 
			res =  gain/(-eigenvalue) + exp(eigenvalue*t)*(1 - gain/(-eigenvalue));  
			break;
	}
	
	if (!pijt_is_prob_value(res)){
		string err = "Error in twoStateModel::Pij_t, pijt <0 or >1. val= " + double2string(res);
		err+=" d=";
		err+= double2string(t);		
		LOG(4,<<err<<endl);		//errorMsg::reportError(err);
	}
	if(!(res>VERYSMALL))
		res = VERYSMALL;
	return res; 
}

/********************************************************************************************
dPij_t - Based on Analytic solution
*********************************************************************************************/
const MDOUBLE twoStateModel::dPij_dt(const int i,const int j, const MDOUBLE t1)  const {
	//MDOUBLE t = t1 * _rQ;
	MDOUBLE t = t1;;
	MDOUBLE gain = _Q[0][1];
	MDOUBLE loss = _Q[1][0];
	MDOUBLE eigenvalue =  -(gain + loss);

	MDOUBLE res = -1;
	int caseNum = i + j*2;
	switch (caseNum) {
		case 0 : 
			res =    exp(eigenvalue*t)*(eigenvalue + loss); 
			break;
		case 1 : 
			res =  -(exp(eigenvalue*t)*(eigenvalue + gain)); //[1][0]
			break;
		case 2 :
			res =  -(exp(eigenvalue*t)*(eigenvalue + loss));  
			break;
		case 3 : 
			res =    exp(eigenvalue*t)*(eigenvalue + gain);  
			break;
	}
	
	return res; 
}
/********************************************************************************************
d2Pij_dt2 - Based on Analytic solution
*********************************************************************************************/
const MDOUBLE twoStateModel::d2Pij_dt2(const int i,const int j, const MDOUBLE t)  const {
	MDOUBLE gain = _Q[0][1];
	MDOUBLE loss = _Q[1][0];
	MDOUBLE eigenvalue =  -(gain + loss);

	MDOUBLE res = -1;
	int caseNum = i + j*2;
	switch (caseNum) {
		case 0 : 
			res =    exp(eigenvalue*t)*(eigenvalue + loss)*eigenvalue; 
			break;
		case 1 : 
			res =  -(exp(eigenvalue*t)*(eigenvalue + gain))*eigenvalue; 
			break;
		case 2 : 
			res =  -(exp(eigenvalue*t)*(eigenvalue + loss))*eigenvalue;  
			break;
		case 3 : 
			res =    exp(eigenvalue*t)*(eigenvalue + gain)*eigenvalue;  
			break;
	}
	return res;
}


bool twoStateModel::pijt_is_prob_value(MDOUBLE val) const 
{
    if ((val+err_allow_for_pijt_function()<0) || (val>1+err_allow_for_pijt_function()))
        return false;
    else
        return true;
}
